/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.analysis;

import jalview.api.analysis.ScoreModelI;
import jalview.api.analysis.SimilarityParamsI;
import jalview.bin.Console;
import jalview.datamodel.AlignmentView;
import jalview.datamodel.Point;
import jalview.datamodel.SequenceI;
import jalview.gui.PairwiseAlignPanel;
import jalview.gui.PaSiMapPanel;
import jalview.math.MatrixI;
import jalview.viewmodel.AlignmentViewport;


import java.io.PrintStream;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * Performs Principal Component Analysis on given sequences
 * @AUTHOR MorellThomas 
 */
public class PaSiMap implements Runnable
{
  /*
   * inputs
   */
  final private AlignmentViewport seqs;

  final private ScoreModelI scoreModel;

  final private SimilarityParamsI similarityParams;

  final private byte dim = 3;

  /*
   * outputs
   */
  private MatrixI pairwiseScores;

  private MatrixI tridiagonal;

  private MatrixI eigenMatrix;

  /**
   * Constructor given the sequences to compute for, the similarity model to
   * use, and a set of parameters for sequence comparison
   * 
   * @param sequences
   * @param sm
   * @param options
   */
  //public PaSiMap(AlignmentView sequences, ScoreModelI sm,
  //&! viewport or panel?
  public PaSiMap(AlignmentViewport sequences, ScoreModelI sm,
          SimilarityParamsI options)
  {
    this.seqs = sequences;
    this.scoreModel = sm;
    this.similarityParams = options;
  }

  /**
   * Returns Eigenvalue
   * 
   * @param i
   *          Index of diagonal within matrix
   * 
   * @return Returns value of diagonal from matrix
   */
  public double getEigenvalue(int i)
  {
    return eigenMatrix.getD()[i];
  }

  /**
   * DOCUMENT ME!
   * 
   * @param l
   *          DOCUMENT ME!
   * @param n
   *          DOCUMENT ME!
   * @param mm
   *          DOCUMENT ME!
   * @param factor
   *          DOCUMENT ME!
   * 
   * @return DOCUMENT ME!
   */
  public Point[] getComponents(int l, int n, int mm, float factor)
  {
    Point[] out = new Point[getHeight()];

    for (int i = 0; i < getHeight(); i++)
    {
      float x = (float) component(i, l) * factor;
      float y = (float) component(i, n) * factor;
      float z = (float) component(i, mm) * factor;
      out[i] = new Point(x, y, z);
    }

    return out;
  }

  /**
   * DOCUMENT ME!
   * 
   * @param n
   *          DOCUMENT ME!
   * 
   * @return DOCUMENT ME!
   */
  public double[] component(int n)
  {
    // n = index of eigenvector
    double[] out = new double[getHeight()];

    for (int i = 0; i < out.length; i++)
    {
      out[i] = component(i, n);
    }

    return out;
  }

  /**
   * DOCUMENT ME!
   * 
   * @param row
   *          DOCUMENT ME!
   * @param n
   *          DOCUMENT ME!
   * 
   * @return DOCUMENT ME!
   */
  double component(int row, int n)
  {
    double out = 0.0;

    for (int i = 0; i < pairwiseScores.width(); i++)
    {
      out += (pairwiseScores.getValue(row, i) * eigenMatrix.getValue(i, n));
    }

    return out / eigenMatrix.getD()[n];
  }

  /**
   * Answers a formatted text report of the PaSiMap calculation results (matrices
   * and eigenvalues) suitable for display
   * 
   * @return
   */
  public String getDetails()
  {
    StringBuilder sb = new StringBuilder(1024);
    sb.append("PaSiMap calculation using ").append(scoreModel.getName())
            .append(" sequence similarity matrix\n========\n\n");
    PrintStream ps = wrapOutputBuffer(sb);

    /*
     * pairwise similarity scores
     */
    sb.append(" --- OrigT * Orig ---- \n");
    pairwiseScores.print(ps, "%8.2f");

    /*
     * tridiagonal matrix, with D and E vectors
     */
    sb.append(" ---Tridiag transform matrix ---\n");
    sb.append(" --- D vector ---\n");
    tridiagonal.printD(ps, "%15.4e");
    ps.println();
    sb.append("--- E vector ---\n");
    tridiagonal.printE(ps, "%15.4e");
    ps.println();

    /*
     * eigenvalues matrix, with D vector
     */
    sb.append(" --- New diagonalization matrix ---\n");
    eigenMatrix.print(ps, "%8.2f");
    sb.append(" --- Eigenvalues ---\n");
    eigenMatrix.printD(ps, "%15.4e");
    ps.println();

    return sb.toString();
  }

  /**
   * Performs the PaSiMap calculation
   */
  @Override
  public void run()
  {
    try
    {
      /**	here goes pasimap
      *	FASTA_to_secureFASTA	+ sed s/ /_/g ~~ aliases = 1st word of secure fasta header
      * FASTA_to_stats	~~ unique_headers_count, unique_bodies_count + check: unique_bodies_min, dim_max
      * if unaligned : EMBOSS needlall ~~ alignments.fas
      * elseif aligned : MSA_to_pairwiseFASTA	
      * pairwiseFASTA_to_pairwiseCSV
      * pairwise_to_connectivity ~~ connectivity.csv	+ warn if sequence has #connections >= dim
      * pairwiseCSV_to_pairwiseQuantifier (use scoreModel)	+ label results
      * cc_analysis quantifier.ssv ~~ vec.ssv	+ label results
      * plot
      *
      * ######################
      *
      * input AlignemtView seqs ~~ check if aligned
      * if not ~~ perform needlall (or try whatever is in there)
      * calculate connectivity + check if fine
      * quantify alignment with ScoreModelI scoreModel
      * cc_analysis with quantification
      * plot whatever comes out of cc_analysis
      */

      /*
      bool aligned = seqs.checkoridk;
      if (!aligned)
      {
	seqs = needlall(seqs);
      }
      
      connectivity

      pairwiseScores = scoreModel.findSimilarities(seqs, similarityParams);

      cc_analysis
      */

      // run needleman regardless if aligned or not
      // gui.PairwiseAlignPanel <++>
      PairwiseAlignPanel alignment = new PairwiseAlignPanel(seqs);
      float[][] scores = alignment.getScores();
      System.out.println(scores[8][18]);	// do scores have to be divided by totscore??

      // what is connectivity and why?
      Hashtable<SequenceI, Integer> connectivity = seqs.calculateConnectivity(scores, dim);


      /** pca code */
      /*
       * sequence pairwise similarity scores
       */
      //&! analysis/scoremodels/ScoreMatrix	computeSimilarity
      //pairwiseScores = scoreModel.findSimilarities(seqs, similarityParams);

      /*
       * tridiagonal matrix
       */
      //tridiagonal = pairwiseScores.copy();
      //tridiagonal.tred();

      /*
       * the diagonalization matrix
       */
      //eigenMatrix = tridiagonal.copy();
      //eigenMatrix.tqli();
    } catch (Exception q)
    {
      Console.error("Error computing PaSiMap:  " + q.getMessage());
      q.printStackTrace();
    }
  }

  /**
   * Returns a PrintStream that wraps (appends its output to) the given
   * StringBuilder
   * 
   * @param sb
   * @return
   */
  protected PrintStream wrapOutputBuffer(StringBuilder sb)
  {
    PrintStream ps = new PrintStream(System.out)
    {
      @Override
      public void print(String x)
      {
        sb.append(x);
      }

      @Override
      public void println()
      {
        sb.append("\n");
      }
    };
    return ps;
  }

  /**
   * Answers the N dimensions of the NxN PaSiMap matrix. This is the number of
   * sequences involved in the pairwise score calculation.
   * 
   * @return
   */
  public int getHeight()
  {
    // TODO can any of seqs[] be null?
    return pairwiseScores.height();// seqs.getSequences().length;
  }

  /**
   * Answers the sequence pairwise similarity scores which were the first step
   * of the PaSiMap calculation
   * 
   * @return
   */
  public MatrixI getPairwiseScores()
  {
    return pairwiseScores;
  }

  public void setPairwiseScores(MatrixI m)
  {
    pairwiseScores = m;
  }

  public MatrixI getEigenmatrix()
  {
    return eigenMatrix;
  }

  public void setEigenmatrix(MatrixI m)
  {
    eigenMatrix = m;
  }

  public MatrixI getTridiagonal()
  {
    return tridiagonal;
  }

  public void setTridiagonal(MatrixI tridiagonal)
  {
    this.tridiagonal = tridiagonal;
  }
}
